#include "Hashtable.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

void Hashtable::construct(int zinitialCapacity, float zloadFactor) {
	checksum1 = rand();
	if(checksum1 == 0 || checksum1 == IS_VALID) checksum1 += 1;
	checksum2 = checksum1 ^ IS_VALID;
	validKey = IS_VALID;
//	printf("checksums: %d and %d, validKey = %d\n", checksum1, checksum2, validKey);
//	sleep(1);
	count=0;
	if (zinitialCapacity <= 0) zinitialCapacity = 101;
	if (zloadFactor <= 0.0) zloadFactor = 0.75;
	loadFactor = zloadFactor;
	tableLength = zinitialCapacity;
	//old table = new (HashtableEntry*[tableLength]);
	table = (HashtableEntry **)malloc(tableLength * sizeof(HashtableEntry*));
	for(int i=0;i<tableLength;i++) table[i] = null;
	threshold = (int)(tableLength * loadFactor);
/*
	printf("Locations of checksums: %d and %d\n", &checksum1, &checksum2);
	printf("Other things: %d, %d, %d, %d, %d\n", &(validKey == IS_VALID), &table, &enumEntry, &count, &tableLength);	
	printf("Other things: %d, %d, %d\n", &enumIndex, &threshold, &loadFactor);	
	sleep(1);
*/
}

Hashtable::Hashtable(int initialCapacity, float myLoadFactor) {
	construct(initialCapacity, myLoadFactor);
}

Hashtable::Hashtable(int initialCapacity) {
	construct(initialCapacity, 0.75f);
}

Hashtable::Hashtable() {
	construct(17, 0.75f);
}

Hashtable::~Hashtable() {
	if(checkValid()) {
		clear();
		free(table);
//old		delete [] table;
	}
}

bool Hashtable::checkValid() {
	// once bad, always bad
	if(validKey == NOT_VALID) return false;
	
	unsigned char sum = checksum1 ^ checksum2;
	if(sum != IS_VALID) {
		validKey = NOT_VALID;
		printf("WARNING: hashtable %X went bad, there is a memory problem\n", (unsigned int)this);
		printf("  Checksum: %x ^ %x == %x != %x)\n", checksum1, checksum2, sum, IS_VALID);
		printf("  size: %d, tableLength: %d, threshhold: %d, loadFactor: %g\n", count, tableLength, threshold, loadFactor);
		return false;
	}
	return true;
}

bool Hashtable::checkEntryValid(HashtableEntry *e) {
	// once bad, always bad
	if(e->validKey == NOT_VALID) return false;
	unsigned char sum = e->checksum1 ^ e->checksum2;
	if((sum != IS_VALID) || e->key == null || e->value == null) {
		e->validKey = NOT_VALID;
		validKey = NOT_VALID;
		printf("WARNING: hashtable entry %X in hashtable %X went bad, there is a memory problem\n", (unsigned int)e, (unsigned int)this);
		printf("  Checksum: %x ^ %x == %x ?= %x)\n", e->checksum1, e->checksum2, sum, IS_VALID);
		printf("  key = %x, value = %x\n", (unsigned int)e->key, (unsigned int)e->value);
		return false;
	}
	return true;
}

/**
 * Returns the number of keys in this hashtable.
 */
int Hashtable::size() {
	if(!checkValid()) return 0;
	else return count;
}

/**
 * Tests if this hashtable maps no keys to values.
 */
bool Hashtable::isEmpty() {
	if(!checkValid()) return true;
	else return count == 0;
}

/**
 * Tests if some key maps into the specified value in this hashtable.
 * This operation is more expensive than the <code>containsKey</code>
 * method.
 */
bool Hashtable::contains(void *value) {
	if(!checkValid()) return false;
	if (value == null) return false;
	for(int i=0;i<tableLength;i++) {
		for (HashtableEntry *e = table[i] ; e != null ; e = e->next) {
			if(!checkEntryValid(e)) return false;
			if (e->value == value) {
	    		return true;
			}
		}
	}
	return false;
}

/**
 * Tests if the specified object is a key in this hashtable.
 */
bool Hashtable::containsKey(const char *key) {
	//HashtableEntry **tab = table;
	if(!checkValid()) return false;
	if(key == null) return false;
	int hash = hashCode(key);
	int ix = (hash & 0x7FFFFFFF) % tableLength;
	for (HashtableEntry *e = table[ix] ; e != null ; e = e->next) {
		if(!checkEntryValid(e)) return false;
		if ((e->hash == hash) && (strcmp(e->key,key) == 0)) {
			return true;
		}
	}
	return false;
}

/**
 * Returns the value to which the specified key is mapped in this hashtable.
 */
void* Hashtable::get(const char *key) {
	//HashtableEntry **tab = table;
	if(!checkValid()) return null;
	if(key == null) return null;
	int hash = hashCode(key);
	int ix = (hash & 0x7FFFFFFF) % tableLength;
	for (HashtableEntry *e = table[ix] ; e != null ; e = e->next) {
		if(!checkEntryValid(e)) return null;
		if ((e->hash == hash) && (strcmp(e->key,key) == 0)) {
				return e->value;
		}
	}
	return null;
}

/**
 * Rehashes the contents of the hashtable into a hashtable with a 
 * larger capacity. This method is called automatically when the 
 * number of keys in the hashtable exceeds this hashtable's capacity 
 * and load factor. 
 */
void Hashtable::rehash() {
	int i;
	int oldCapacity = tableLength;
	HashtableEntry **oldTable = table;

	int newCapacity = oldCapacity * 2 + 1;
	HashtableEntry **newTable = (HashtableEntry **)malloc(newCapacity * sizeof(HashtableEntry*));
	//oldHashtableEntry **newTable = new (HashtableEntry*[newCapacity]);
	for(i=0;i<newCapacity;i++) newTable[i] = null;

//	printf("rehashing %d to %d\n", oldCapacity, newCapacity);

	threshold = (int)(newCapacity * loadFactor);
	table = newTable;
	tableLength = newCapacity;
	
	//System.out.println("rehash old=" + oldCapacity + ", new=" + newCapacity + ", thresh=" + threshold + ", count=" + count);

//	for (i = oldCapacity ; i-- > 0 ;) {
	for(i=0;i<oldCapacity;i++) {
		for (HashtableEntry *old = oldTable[i] ; old != null ; ) {
			HashtableEntry *e = old;
			old = old->next;

			int ix = (e->hash & 0x7FFFFFFF) % newCapacity;
			e->next = newTable[ix];
			newTable[ix] = e;
	    }
	}
	free(oldTable);
	//old delete [] oldTable;
	//TODO: free oldTable and each oldTable entry?
}

/**
 * Maps the specified <code>key</code> to the specified 
 * <code>value</code> in this hashtable. Neither the key nor the 
 * value can be <code>null</code>. 
 * <p>
 * The value can be retrieved by calling the <code>get</code> method 
 * with a key that is equal to the original key. 
 *
 * returns null or old value for that key
 */
void* Hashtable::put(const char *key, void *value) {
	HashtableEntry *e;
	// Make sure the value is not null
	if(!checkValid()) return null;
	if(key == null) return null;
	if(value == null) return null;

	// Makes sure the key is not already in the hashtable.
	//HashtableEntry **tab = table;
	int hash = hashCode(key);

	int ix = (hash & 0x7FFFFFFF) % tableLength;
	for (e = table[ix] ; e != null ; e = e->next) {
		if(!checkEntryValid(e)) return null;
		if ((e->hash == hash) && (strcmp(e->key,key) == 0)) {
			void *old = e->value;
			e->value = value;
			return old;
		}
	}

	if (count >= threshold) {
		// Rehash the table if the threshold is exceeded
		rehash();
		return put(key, value);
	} 

	// Creates the new entry.
	e = new HashtableEntry();
	e->hash = hash;
	e->key = new char[strlen(key)+1];
//	e->key = (char *)malloc (40 + strlen (key));
	strcpy (e->key, key);
	e->value = value;
	e->next = table[ix];
	table[ix] = e;
	count++;
	return null;
}

int Hashtable::hashCode(const char* key) {
	unsigned long h = 0;
	while(*key) {
		h = (h << 4) + *key++;
		unsigned long g = h & 0xF0000000L;
		if (g) h ^= g >> 24;
		h &= ~g;
	}
	return h;
	//return h % count;
}

/**
 *
 * This method removes the key.It's up to you to delete or free the returned value
 *
 * Removes the key (and its corresponding value) from this 
 * hashtable. This method does nothing if the key is not in the hashtable.
 *
 * @param   key   the key that needs to be removed.
 * @return  the value to which the key had been mapped in this hashtable,
 *          or <code>null</code> if the key did not have a mapping.
 * @since   JDK1.0
 */
void* Hashtable::remove(char *key) {
	if(!checkValid()) return null;
	if(key == null) return null;
	
	int hash = hashCode(key);
	int ix = (hash & 0x7FFFFFFF) % tableLength;
	void *retval;
	for (HashtableEntry *e = table[ix], *prev = null ; e != null ; prev = e, e = e->next) {
		if(!checkEntryValid(e)) return null;
		if ((e->hash == hash) && (strcmp(e->key, key) == 0)) {
			if (prev != null) {
				prev->next = e->next;
			} 
			else {
				table[ix] = e->next;
			}
			count--;
			retval = e->value;
			delete e;
			return retval;
//			return e->value;
		}
	}
	return null;
}

/**
 * Clears this hashtable so that it contains no keys. 
 */
void Hashtable::clear() {
	int i;
	HashtableEntry *trash;

	if(!checkValid()) return;
	for(i=0;i<tableLength;i++) {
		for (HashtableEntry *e = table[i] ; e != null ; ) {
			if(!checkEntryValid(e)) return;
			trash = e;
			e = e->next;
			delete trash;
		}
		table[i] = null;
	}
	count = 0;
}

/**
 * Clears this hashtable so that it contains no keys. 
 * Also calls free on all stored values
 */
void Hashtable::clearAndFree() {
	int i;
	HashtableEntry *trash;

	if(!checkValid()) return;
	for(i=0;i<tableLength;i++) {
		for (HashtableEntry *e = table[i] ; e != null ; ) {
			if(!checkEntryValid(e)) return;
			trash = e;
			e = e->next;
			free(trash->value);
			delete trash;
		}
		table[i] = null;
	}
	count = 0;
}

    /**
     * Creates a shallow copy of this hashtable. The keys and values 
     * themselves are not cloned. 
     * This is a relatively expensive operation.
     *
     * @return  a clone of the hashtable.
     * @since   JDK1.0
     */
		/* WHY CLONE?
  Object clone() {
	try { 
	    Hashtable t = (Hashtable)super.clone();
	    t.table = new HashtableEntry[table.length];
	    for (int i = table.length ; i-- > 0 ; ) {
		t.table[i] = (table[i] != null) 
		    ? (HashtableEntry)table[i].clone() : null;
	    }
	    return t;
	} catch (CloneNotSupportedException e) { 
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
    }
    */

    /**
     * Returns a rather long string representation of this hashtable.
     *
     * @return  a string representation of this hashtable.
     * @since   JDK1.0
     */
		/* TOSTRING WHAT?
		String toString() {
			int max = size() - 1;
			StringBuffer buf = new StringBuffer();
			Enumeration k = keys();
	Enumeration e = elements();
	buf.append("{");

	for (int i = 0; i <= max; i++) {
	    String s1 = k.nextElement().toString();
	    String s2 = e.nextElement().toString();
	    buf.append(s1 + "=" + s2);
	    if (i < max) {
		buf.append(", ");
	    }
	}
	buf.append("}");
	return buf.toString();
    }
		*/
		
    /**
     * WriteObject is called to save the state of the hashtable to a stream.
     * Only the keys and values are serialized since the hash values may be
     * different when the contents are restored.
     * iterate over the contents and write out the keys and values.
     */
     /* STREAM WHAT?
    private synchronized void writeObject(java.io.ObjectOutputStream s)
        throws IOException
    {
	// Write out the length, threshold, loadfactor
	s.defaultWriteObject();

	// Write out length, count of elements and then the key/value objects
	s.writeInt(table.length);
	s.writeInt(count);
	for (int ix = table.length-1; ix >= 0; ix--) {
	    HashtableEntry entry = table[ix];

	    while (entry != null) {
		s.writeObject(entry.key);
		s.writeObject(entry.value);
		entry = entry.next;
	    }
	}
    }
		*/
		
    /**
     * readObject is called to restore the state of the hashtable from
     * a stream.  Only the keys and values are serialized since the
     * hash values may be different when the contents are restored.
     * Read count elements and insert into the hashtable. 
     */
		/* STREAM WHAT?
    private synchronized void readObject(java.io.ObjectInputStream s)
         throws IOException, ClassNotFoundException
    {
	// Read in the length, threshold, and loadfactor
	s.defaultReadObject();

	// Read the original length of the array and number of elements
	int origlength = s.readInt();
	int elements = s.readInt();

	// Compute new size with a bit of room 5% to grow but
	// No larger than the original size.  Make the length
	// odd if it's large enough, this helps distribute the entries.
	// Guard against the length ending up zero, that's not valid.
	int length = (int)(elements * loadFactor) + (elements / 20) + 3;
	if (length > elements && (length & 1) == 0)
	    length--;
	if (origlength > 0 && length > origlength)
	    length = origlength;

	table = new HashtableEntry[length];
	count = 0;

	// Read the number of elements and then all the key/value objects
	for (; elements > 0; elements--) {
	    Object key = s.readObject();
	    Object value = s.readObject();
	    put(key, value);
	}
    }
}
*/

/**
 * A hashtable enumerator class.  This class should remain opaque 
 * to the client. It will use the Enumeration interface. 
 */
    
void Hashtable::resetEnumerator(void) {
	enumIndex = -1;
	enumEntry = null;
}

bool Hashtable::hasMoreKeys() {
	if(!checkValid()) return false;

	if(enumEntry != null) {
		return true;
	}

	++enumIndex;
	while (enumIndex < tableLength) {
		if((enumEntry = table[enumIndex]) != null) {
			return true;
		}
		++enumIndex;
	}

	return false;
}

char* Hashtable::nextKey() {
	char *retVal;

	if(!checkValid()) return null;

	// scan upwards if current entry is null	
	while (enumEntry == null && enumIndex < tableLength-1) {
		// check in another table slot
		++enumIndex;
		enumEntry = table[enumIndex];
	}
	if(enumEntry != null) {
		retVal = enumEntry->key;
		enumEntry = enumEntry->next;
		return retVal;
	}
	else 
		return null;
}

	/* ENUMERATE WHAT?
class HashtableEnumerator : public Enumeration {
    boolean keys;
    int index;
    HashtableEntry table[];
    HashtableEntry entry;

    HashtableEnumerator(HashtableEntry table[], boolean keys) {
	this.table = table;
	this.keys = keys;
	this.index = table.length;
    }
	
    public boolean hasMoreElements() {
	if (entry != null) {
	    return true;
	}
	while (index-- > 0) {
	    if ((entry = table[index]) != null) {
		return true;
	    }
	}
	return false;
    }

    public Object nextElement() {
	if (entry == null) {
	    while ((index-- > 0) && ((entry = table[index]) == null));
	}
	if (entry != null) {
	    HashtableEntry e = entry;
	    entry = e.next;
	    return keys ? e.key : e.value;
	}
	throw new NoSuchElementException("HashtableEnumerator");
    }
}
*/
