#ifndef _HASHTABLE_H_
#define _HASHTABLE_H_

#include <string.h>
#include <stdlib.h>

#ifndef null
#define null NULL
#endif

#define IS_VALID 0xA5
#define NOT_VALID ((unsigned char)~IS_VALID)

/*
 * This hashtable greatly resembles the java one
 */

/**
 * Hashtable collision list.
 */
class HashtableEntry {
	public:
		unsigned char checksum1;
		unsigned char validKey;
		char *key;
		void *value;
		int hash;
		HashtableEntry *next;
		unsigned char checksum2;

		HashtableEntry () { 
			checksum1 = rand();
			if(checksum1 == 0 || checksum1 == IS_VALID) checksum1 += 1;
			checksum2 = checksum1 ^ IS_VALID;
			validKey = IS_VALID;
			key = NULL;
			value = NULL;
			hash = -1;
			next = NULL;
		}

		~HashtableEntry () { 
//			if(key!=NULL) printf("deleting %d blocks\n", strlen(key));
			if (key!=NULL) delete [] key;
//			if (key!=NULL) free(key);
//			free ((void *)key);
		}

	protected:
/*
		void* clone() {
			HashtableEntry *entry = new HashtableEntry();
			entry->hash = hash;
			entry->key = key;
			entry->value = value;
			entry->next = (next != null) ? (HashtableEntry *)next->clone() : null;
			return (void *)entry;
		}
*/
};

class Hashtable {
private:
	unsigned char checksum1;
	unsigned char validKey;
	HashtableEntry **table, *enumEntry;
	int count;
	int tableLength, enumIndex;
	int threshold;
	float loadFactor;
	void construct(int zinitialCapacity, float zloadFactor);
	void rehash();
	int hashCode(const char* key);

public:
	Hashtable(int initialCapacity, float loadFactor);
	Hashtable(int initialCapacity);
	Hashtable();
	~Hashtable();
	int size();
	bool isEmpty();
	bool contains(void *value);
	bool containsKey(const char *key);
	void* get(const char *key);
	void* put(const char *key, void *value);
	void* remove(char *key);
	void clear();
	void clearAndFree();
	void resetEnumerator(void);
	bool hasMoreKeys();
	char* nextKey();
	bool checkValid();
	bool checkEntryValid(HashtableEntry *e);
	unsigned char checksum2;
};

#endif
