#include <stdio.h>
#include <math.h>
#include "writehex.h"
#include "gethex.h"
#include "compressMovie.h"
#include "PropertyList.h"

using namespace std; 	//introduces namespace std

#define PRINT_FRAMES false
#define PRINT_DECOMP_FRAMES true
#define PRINT_WRITE_BITS false
#define PRINT_ON_DECODE true


#define ONE_THRESH 85
#define TWO_THRESH 170

// Colors
#define OFF  0
#define HALF 1
#define ON   2

#define W HOTPANTS_W
#define H HOTPANTS_H
	
	int movieError = MOVIE_NO_ERROR;

	bool outBuff[MAX_BITS];
	unsigned char MOVIE[(MAX_BITS/8)+1];
	unsigned int readBitCounter;
	long  bitsUsed = 0;
	int  nowState[W][H];
	int  nextChange[W][H];
	int  framesEncoded = 0;

	unsigned short movieBytePtr;
	char  movieBitPtr;
	bool compressed = false;
	PropertyList*  prefs;


int countFileFrames(char* baseFile, char* fileSuffix) {
	bool done = false;
	char fileName[255];
	int count = 0;
	FILE* file;
	
	while (!done) {
		count++;

		// Up to 999 frames as 001, 002, etc...
	  if (count < 10)
			sprintf(fileName, "%s00%1d%s", baseFile, count, fileSuffix);
		else if (count < 100)
			sprintf(fileName, "%s0%2d%s", baseFile, count, fileSuffix);
		else
			sprintf(fileName, "%s%3d%s", baseFile, count, fileSuffix);

		file = fopen(fileName, "r");
		if (!file || count > MAX_FRAMES) {
			return count - 1;
		}
		fclose(file);
	}
	return 0;
}



void readFrames(char* baseFile, char* fileSuffix, unsigned char frameBuff[][W][H], int& frames) {
	char fileName[255];
	int nFrame = 0;
	unsigned char byte;
	FILE* file;
	
	frames = MIN(MAX_FRAMES, frames);
	
	for (int nFrame = 0; nFrame < frames; nFrame++) {

	  if (nFrame < 9)
			sprintf(fileName, "%s00%1d%s", baseFile, nFrame+1, fileSuffix);
		else if (nFrame < 99)
			sprintf(fileName, "%s0%2d%s", baseFile, nFrame+1, fileSuffix);
		else
			sprintf(fileName, "%s%3d%s", baseFile, nFrame+1, fileSuffix);

		file = fopen(fileName, "r");
		if (!file) {
			frames = nFrame;
			return;
		}
		//printf("Processing file %s\n", fileName);
		
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++) {
				byte = fgetc(file);
				if (byte > TWO_THRESH) {
					frameBuff[nFrame][x][y] = 2;
					if (PRINT_FRAMES) printf("*");
				}
				else if (byte > ONE_THRESH) {
					frameBuff[nFrame][x][y] = 1;
					if (PRINT_FRAMES) printf(".");
				}
				else {
					frameBuff[nFrame][x][y] = 0;
					if (PRINT_FRAMES) printf(" ");
				}
			}
			if (PRINT_FRAMES) printf("\n");
		}
		if (PRINT_FRAMES) printf("\n");
		
		fclose(file);
	}

}

void setMovieBit(int bitNum, bool value) {
	int movieByte = bitNum / 8;
	unsigned char bitOffset = 7 - (bitNum % 8);
	
	if (value) {
	  MOVIE[movieByte] |= ((char)1<<bitOffset);
	}
	else {
		MOVIE[movieByte] &= ~((char)1<<bitOffset);
	}
}

bool readBit() {
	return outBuff[readBitCounter++];
}


bool writeBits(int value, int nBits) {
	if (bitsUsed + nBits >= MAX_BITS) {
		movieError = MOVIE_TOO_BIG_ERROR;
		return false;
	}

	int maxVal = pow(2, nBits) - 1;
	if (value > maxVal) {
		//printf("Write bits error. %d bits are not enough to encode %d.\n", nBits, value);
		movieError = MOVIE_GENERAL_ERROR;
		return false;
	}
	
	
	for (int i = 0; i < nBits; i++) {
		outBuff[bitsUsed + i] = false;
		setMovieBit(bitsUsed + i, false);
	}

	int divider;
	int remainder = value;	
	for (int i = nBits - 1; i >= 0; i--) {
		divider = pow(2, i);
		if (remainder >= divider) {
			remainder -= divider;
			outBuff[bitsUsed + (nBits - (i+1))] = true;
			setMovieBit(bitsUsed + (nBits - (i+1)), true);
		}
	}
		
	if (PRINT_WRITE_BITS) {
		printf("Bit addr: %d, Bin val: ", value);
		for (int i = 0; i < nBits; i++) {
			printf("%d ", outBuff[bitsUsed + i]);
		}
		printf("\n");
	}
	
	bitsUsed += nBits;
	return true;
}

int bitsNeeded(int value) {
  int nBits = 1;
 	while (value > pow(2, nBits) - 1)
 		nBits++;
 	return nBits;
}

// Will write the value in the smallest number of bits
bool writeBits(int value) {
  int nBits = 1;
 	while (value > pow(2, nBits) - 1)
 		nBits++;
 	return writeBits(value, nBits);
}


bool willFit() {
	
	return true;
}

// This just gives a one-bit bootstrap so that from now on state change can be encoded in one bit
void resetCompress() {
	movieError = MOVIE_NO_ERROR;
	compressed = false;
  bitsUsed = 0;
  framesEncoded = 0;
  movieBytePtr = 0;
  movieBitPtr = 7;
  
	for (int y = 0; y < H; y++) {
		for (int x = 0; x < W; x++) {
			nextChange[x][y] = 0;
		}
	}
}

int tillNext(int start, int x, int y, unsigned char frameBuff[][W][H], int frames) {
	int till = 0;
	int index = start;
	
	while (index < frames && frameBuff[index][x][y] == nowState[x][y]) {
		index++;
		till++;
	}
	return till;
}


void compressNBitChange(int changeBits, unsigned char frameBuff[][W][H], int frames) {	
	int next;
	int bits;
	int metaBits = bitsNeeded(changeBits-1);
	
	
	resetCompress();
	compressed = true;

	int nFrame;
	int topFrame = MIN(MAX_FRAMES, frames);
	
	for (nFrame = 0; nFrame < topFrame; nFrame++) {
		framesEncoded = nFrame;
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++) {
				if (nextChange[x][y] == nFrame) {				
					if (!writeBits(frameBuff[nFrame][x][y], 2)) {
						return;
					}
					nowState[x][y] = frameBuff[nFrame][x][y];
					
					next = tillNext(nFrame+1, x, y, frameBuff, frames);
		      bits = bitsNeeded(next);
					if (bits > changeBits) {
						next = pow(2,changeBits) - 1;
						bits = changeBits;
					}
					if (!writeBits(bits-1, metaBits))
						return;
					if (!writeBits(next))
						return;
					nextChange[x][y] = nFrame + next + 1;
				}
			}
		}
	}
	framesEncoded = topFrame;
}



void compressNBitRun(int runLengthBits, unsigned char frameBuff[][W][H], int frames) {	
	int curPixel = frameBuff[0][0][0];
	int runState = curPixel + 1;
	int runLength = 1;
	bool starting = true;
	
	resetCompress();
	compressed = true;
	int nFrame;
	int topFrame = MIN(MAX_FRAMES, frames);

	for (nFrame = 0; nFrame < topFrame; nFrame++) {
		framesEncoded = nFrame;
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++) {
				curPixel = frameBuff[nFrame][x][y];
				if (curPixel != runState || bitsNeeded(runLength) > runLengthBits) {
					if (starting) {
						starting = false;
					}
					else {
						if (!writeBits(runLength - 1, runLengthBits))
							return;
					}
					runState = curPixel;
					runLength = 1;
					if (!writeBits(runState, 2))
						return;
				}
				else {
					runLength++;
				}
			}
		}
	}
	writeBits(runLength - 1, runLengthBits);
	framesEncoded = topFrame;
}


int diff(int from, int to) {
	int aFrom;
	
	if (from > 1)
		aFrom = -1;
	else
		aFrom = from;

	return to - aFrom;	
}



void compressMarkChange(unsigned char frameBuff[][W][H], int frames) {	
	int curPixel;
	int lastState = 0;
	
	resetCompress();
	int nFrame;
	int topFrame = MIN(MAX_FRAMES, frames);

	for (nFrame = 0; nFrame < topFrame; nFrame++) {
		framesEncoded = nFrame;
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++) {
				curPixel = frameBuff[nFrame][x][y];
				if (curPixel == lastState) {
					if (!writeBits(0, 1))
						return;
				}
				else if (diff(lastState, curPixel) == 1) {
					if (!writeBits(2, 2))
						return;
				}
				else {
					if (!writeBits(3, 2))
						return;
				}
				lastState = curPixel;
			}
		}
	}
	framesEncoded = topFrame;
	compressed = true;
}


void printReport() {
	printf("%d bits = %4.2f bytes used for %d frames.\n", bitsUsed, bitsUsed / 8.0f, framesEncoded);
	printf("Average of %4.2f bits = %4.2f bytes per frame.\n", (float)bitsUsed/ framesEncoded, (bitsUsed / 8.0f) / framesEncoded);

}

//===================== DECODE
unsigned char delayCounter;

unsigned char  curMovieData;

unsigned char nowDecodeState;
unsigned char curFrame;

void initMovie() {
  readBitCounter = 0;

  nowDecodeState = 0;

	curFrame = 0;
	movieBytePtr = 0;
	movieBitPtr = 7;
	curMovieData = MOVIE[0];
}

unsigned char getBits(char nBits) {
	char i;
	unsigned char result = 0;
	
	for (i = 0; i < nBits; i++) {
		result <<= 1;
		if (curMovieData & (1 << movieBitPtr))
			result++;
		movieBitPtr--;
		if (movieBitPtr < 0) {
			movieBitPtr = 7;
			movieBytePtr++;
			curMovieData = MOVIE[movieBytePtr];			
		}
	}
	return result;
}

void nextFrame(int frames) {
	char x, y;
	char nextBit;
	
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			nextBit = getBits(1);
			if (nextBit) {
				nextBit = getBits(1);
				if (!nextBit) {
					nowDecodeState = (nowDecodeState + 1) % 3;
				}
				else {
					nowDecodeState = (nowDecodeState + 2) % 3;
				}
			}

			if(PRINT_ON_DECODE) {
				if (nowDecodeState == 0)
					printf(" ");
				else if (nowDecodeState == 1)
					printf(".");
				else
					printf("*");
			}
		}
		if(PRINT_ON_DECODE) printf("\n");
	}
	if(PRINT_ON_DECODE) printf("\n");
	
	curFrame++;
	if (curFrame == frames) {
		initMovie();
	}
}



void decompressMarkChange(int frames) {
	initMovie();
	for (int i = 0; i < frames; i++) {
		nextFrame(frames);
	}
}

// Decompress NBitChange...
unsigned long nowState1[10];
unsigned long nowState2[10];
unsigned int nextChange1[70];
unsigned int nextChange2[70];
int  nBitChangeOutFrame[W][H];
int META_BITS;

void setNextChange(int x, int y, int value) {
	if (y < 7) {
		nextChange1[y * 10 + x] = value; 
	}
	else {
		nextChange2[(y-7) * 10 + x] = value; 
	}
}


int getNextChange(int x, int y) {
	if (y < 7) {
		return nextChange1[y * 10 + x]; 
	}
	else {
		return nextChange2[(y-7) * 10 + x]; 
	}
}

void initNBitChangeMovie() {
  readBitCounter = 0;

  nowDecodeState = 0;
  
	for (int y = 0; y < H; y++) {
		for (int x = 0; x < W; x++) {
			//nextChange[x][y] = 0;
			setNextChange(x, y, 0);
		}
	}

	curFrame = 0;
	movieBytePtr = 0;
	movieBitPtr = 7;
	curMovieData = MOVIE[0];
}


void setNowState(int x, int y, int value) {
  if (value == OFF) {
    nowState1[x] &= ~(1L<<(long)y);
    nowState2[x] &= ~(1L<<(long)y);
  }
  else if (value == HALF) {
    nowState1[x] |= (1L<<(long)y);
    nowState2[x] &= ~(1L<<(long)y);
  }
  else {
    nowState1[x] |= (1L<<(long)y);
    nowState2[x] |= (1L<<(long)y);
  }
}


int getNowState(int x, int y) {
	if (nowState2[x] & (1L<<(long)y))
		return ON;

	if (nowState1[x] & (1L<<(long)y))
		return HALF;
	
	return OFF;
}

void point(int x, int y, int state) {
	nBitChangeOutFrame[x][y] = state;
}

void nextNBitChangeFrame() {
	int nextBits;
	int nextVal;
	int x, y;
	
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			if (getNextChange(x,y) == curFrame) {  //nextChange[x][y] == curFrame) {
				//nowState[x][y] = getBits(2);
				setNowState(x,y,getBits(2));
				nextBits = getBits(META_BITS) + 1;
				nextVal = getBits(nextBits);
				//nextChange[x][y] = curFrame + nextVal + 1;
				setNextChange(x, y, curFrame + nextVal + 1);
			}
			
			//point(x, y, nowState[x][y]);
			point(x, y, getNowState(x,y));
		}
	}
	
	curFrame++;
}

void decompressNBitChange(int changeBits, int frames) {
	META_BITS = bitsNeeded(changeBits-1);
	
	
	initNBitChangeMovie();
	int pixel;
	int nFrame;
	for (nFrame = 0; nFrame < frames; nFrame++) {
		nextNBitChangeFrame();
		if (PRINT_DECOMP_FRAMES) printf("Decompressed frame %d\n", nFrame + 1);
		
		for (int y = 0; y < H; y++) {
			for (int x = 0; x < W; x++) {
				pixel = nBitChangeOutFrame[x][y];
				if (pixel == 2) {
					if (PRINT_DECOMP_FRAMES) printf("*");
				}
				else if (pixel == 1) {
					if (PRINT_DECOMP_FRAMES) printf(".");
				}
				else {
					if (PRINT_DECOMP_FRAMES) printf(" ");
				}
			}
			if (PRINT_DECOMP_FRAMES) printf("\n");
		}
		if (PRINT_DECOMP_FRAMES) printf("\n");
	}

}


//========== OUTPUT

bool writeCompressedMovie(char* outFileName, int frameDelay) {
	FILE* outfile;
	int bytesUsed;
	int BREAK_AT = 20;
	int lineBreakCounter = 0;
	
	outfile = fopen(outFileName, "w");
	if (!outfile) {
		printf("Error opening output file %s.\n", outFileName);
		return false;
	}
	
	if (bitsUsed % 8 == 0) {
		bytesUsed = bitsUsed / 8;
	}
	else {
		bytesUsed = (bitsUsed / 8) + 1;
	}
	
	fprintf(outfile, "//Movie data\n");
	fprintf(outfile, "#define FRAME_DELAY %d\n", frameDelay);
	fprintf(outfile, "unsigned long frames = %d;\n", framesEncoded);
	
	int bytesLeft = bytesUsed;
	int bytesInVar;
	int byteStart = 0;
	char movieVarSub[3] = "";
	
	for (int movieVar = 0; movieVar < MAX_MOVIE_VARS; movieVar++) {
	  lineBreakCounter = 0;
	
		if (bytesLeft > 256) {
			bytesInVar = 256;
		}
		else {
			bytesInVar = bytesLeft;
		}
		bytesLeft -= bytesInVar;
		
		movieVarSub[0] = 0;
		sprintf(movieVarSub, "%d%d", movieVar / 10, movieVar % 10);

		if (bytesInVar == 0) {
			fprintf(outfile, "const unsigned int MOVIE%s[1] = {0};\n", movieVarSub);

		}
		else {
		  fprintf(outfile, "const unsigned int MOVIE%s[%d] = {", movieVarSub, bytesInVar);

			for (int i = byteStart; i < byteStart + bytesInVar - 1; i++) {
				fprintf(outfile, "%d, ", MOVIE[i]);
				lineBreakCounter++;
				if (lineBreakCounter > BREAK_AT) {
					lineBreakCounter = 0;
					fprintf(outfile, "\n                                  ");
				}
			}
			fprintf(outfile, "%d};\n", MOVIE[byteStart + bytesInVar - 1]);
		}
		
		byteStart += bytesInVar;
	}
	
	/*
	
	fprintf(outfile, "const unsigned int MOVIE[%d] = {", bytesUsed);
	for (int i = 0; i < bytesUsed - 1; i++) {
		fprintf(outfile, "%d, ", MOVIE[i]);
		lineBreakCounter++;
		if (lineBreakCounter > BREAK_AT) {
			lineBreakCounter = 0;
			fprintf(outfile, "\n                                  ");
		}
	}
	fprintf(outfile, "%d};\n", MOVIE[bytesUsed - 1]);*/

	fclose(outfile);
	return true;
}


bool writeHexFile(char* hexFileName) {
	unsigned int  startAddr = DATA_START_ADDR;

	int nWords;
	if (bitsUsed % 14 == 0)
		nWords = bitsUsed / 14;
	else
		nWords = (bitsUsed / 14) + 1;
	
	nWords++;
	
	unsigned char* buffer = (unsigned char*) malloc(nWords * 2);
	
	int bytePtr = 1;
	int bitPtr = 0;
	
	// write number of frames
	buffer[bytePtr] = 0;
	for (int i = 0; i < 6; i++) {
		buffer[bytePtr] <<= 1;
		if (framesEncoded & (1L << (13 - i)))
			buffer[bytePtr]++;
	}
	bytePtr--;
	buffer[bytePtr] = 0;
	for (int i = 0; i < 8; i++) {
		buffer[bytePtr] <<= 1;
		if (framesEncoded & (1L << (7 - i)))
			buffer[bytePtr]++;
	}
	bytePtr += 3;
	
	for (int word = 1; word < nWords; word++) {
		buffer[bytePtr] = 0;
		for (int bit = 0; bit < 6; bit++) {
			buffer[bytePtr] <<= 1;
			if (bitPtr < bitsUsed)
				buffer[bytePtr] += outBuff[bitPtr];
			bitPtr++;
		}
		bytePtr--;
		buffer[bytePtr] = 0;
		for (int bit = 0; bit < 8; bit++) {
			buffer[bytePtr] <<= 1;
			if (bitPtr < bitsUsed)
				buffer[bytePtr] += outBuff[bitPtr];
			bitPtr++;
		}
		bytePtr += 3;
	}
	writeHex(startAddr, buffer, nWords * 2, hexFileName);
	
	free(buffer);
	return true;
}

int compressMovie(unsigned char frameBuff[][W][H], int frames) {
	compressMarkChange(frameBuff, frames);
	return movieError;
}

int compressToHexFile(char* hexFileName, unsigned char frameBuff[][W][H], int frames) {
	compressMarkChange(frameBuff, frames);
	if (movieError != MOVIE_NO_ERROR)
		return movieError;
	writeHexFile(hexFileName);
	return movieError;
}


bool movieFits() {
	if (!compressed)
		return false;
	return (movieError != MOVIE_TOO_BIG_ERROR);
}

bool movieFits(unsigned char frameBuff[][HOTPANTS_W][HOTPANTS_H], int frames) {
	return (compressMovie(frameBuff, frames) != MOVIE_TOO_BIG_ERROR);
}

int getFramesEncoded() {
	return framesEncoded;
}

long getBitsUsed() {
	return bitsUsed;
}

int main( void )
{
	int  frameDelay = 10;
	int  frames = 0;
	
	char* prefix;
	char* suffix;

	prefs = new PropertyList("compress.pref");
	prefix = prefs->getPropertys("framePrefix");
	suffix = prefs->getPropertys("frameSuffix");

	frames = countFileFrames(prefix, suffix);
	printf("Count file frames: %d\n", frames);
	unsigned char  (*frameBuff)[W][H] = (unsigned char (*)[W][H])malloc(W * H * frames);
	readFrames(prefix, suffix, frameBuff, frames);
	
	printf("Compress by mark change.\n");
	
	compressMovie(frameBuff, frames);
	decompressMarkChange(frames);
	printReport();
	writeHexFile("movie.hex");
	printf("Wrote hex file \"movie.hex\". Frames encoded: %d\n", getFramesEncoded());

	writeCompressedMovie("movie.txt", 15);
	printf("Wrote txt file \"movie.txt\". Frames encoded: %d\n", getFramesEncoded());

	free(frameBuff);

	return 0;
}
