/*
 * Write out the contents of an Intel .hex file as documented at
 *    http://www.piclist.com/techref/fileext/hexs.htm
 *    and http://www.piclist.com/techref/fileext/hex/intel.htm
 * 
 * Note that only the '00' and '01' record formats are used
 */

#include "writehex.h"

#include <stdio.h>
#include <stdlib.h>

#ifndef MIN
#define MIN(x,y) (((x) < (y)) ? (x) : (y))
#endif


// Four bit hex to ASCII
char px(unsigned char h) {
	if(h >= 0xA && h <= 0xF)
		return (h - 10) + 'A';
	return h + '0';
}

// Eight bit hex to ASCII
void p2x(unsigned int h, char *s)
{
	s[0] = px(0xF & (h >> 4));
	s[1] = px(0xF & h);
}

// Sixteen bit hex to ASCII
void p4x(unsigned int h, char *s)
{
	p2x(0xFF & (h >> 8), s);
	p2x(0xFF & h, s+2);
}


unsigned char writeHex(unsigned int startAddr, unsigned char* data, unsigned int length, char *fname) {
  FILE *hexfile;
  char s[4];
  unsigned char cs;
  //unsigned count,addr,rectype,loop;
    
  if ((hexfile = fopen(fname,"w"))==NULL) {
		fprintf(stderr,"Can't create hex file: %s\n",fname);
		return 0;
  }
  
  unsigned int left = length;
  unsigned int addr = startAddr * 2;
  unsigned int ptr = 0;
  
  int recordBytes;
  while (left > 0) {
  	fputc(':',hexfile);
  	recordBytes = MIN(left, 16);
  	cs = recordBytes;
  	p2x(recordBytes, s);
  	fprintf(hexfile, "%2.2s", s);
  	
  	p4x(addr, s);
  	fprintf(hexfile, "%4.4s", s);
  	cs += (0xFF & (addr >> 8));
  	cs += (0xFF & addr);
  	
  	fprintf(hexfile, "00");
  	
  	for (int i = 0; i < recordBytes; i++) {
  		p2x(data[ptr], s);
  		cs += data[ptr];
  		fprintf(hexfile, "%2.2s", s);
  		ptr++;
  	}
  	
  	cs = 0x100 - cs;
  	p2x(cs, s);
  	fprintf(hexfile, "%2.2s", s);
  	
  	fprintf(hexfile, "\n");
  	
  	addr += recordBytes;
  	left -= recordBytes;
  }
  
  fprintf(hexfile, ":00000001FF");
  fprintf(hexfile, "\n");

  
  fclose(hexfile);
  return 0;
}

